@@ -2,7 +2,7 @@
 #define _CALIBRATE_H
 
 /*
- *      Copyright (C) 1993 Bas Laarhoven.
+ *      Copyright (C) 1993,1994 Bas Laarhoven.
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -19,17 +19,18 @@ along with this program; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
  *
- $Source: /usr/src/distr/ftape-0.9.9d/RCS/calibr.h,v $
+ $Source: /home/bas/distr/ftape-1.14d/RCS/calibr.h,v $
  $Author: bas $
  *
- $Revision: 1.8 $
- $Date: 1994/01/16 15:44:54 $
- $State: ALPHA $
+ $Revision: 1.14 $
+ $Date: 1994/06/12 16:07:53 $
+ $State: Beta $
  *
  *      This file contains a gp calibration routine for
  *      hardware dependent timeout functions.
  */
 
+#include <linux/timex.h>
 
 extern int calibrate( char* name, void(*fun)(int), int* calibr_count, int* calibr_time);
 extern void do_gettimeofday( struct timeval *tv);
