@@ -20,39 +20,64 @@
  * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,
  * USA.
  *
- * $Source: /usr/src/distr/ftape-0.9.9d/RCS/ecc.c,v $
+ * $Source: /home/bas/distr/ftape-1.14d/RCS/ecc.c,v $
  * $Author: bas $
  *
- * $Revision: 1.19 $
- * $Date: 1994/01/16 15:44:54 $
- * $State: ALPHA $
+ * $Revision: 1.28 $
+ * $Date: 1994/06/12 16:07:53 $
+ * $State: Beta $
  *
  *      This file contains the Reed-Solomon error correction code 
  *      for the QIC-40/80 floppy-tape driver for Linux.
  */
 static const char RCSid[] =
-  "$Id: ecc.c,v 1.19 1994/01/16 15:44:54 bas ALPHA $";
+  "$Id: ecc.c,v 1.28 1994/06/12 16:07:53 bas Beta $";
 
 #include <stdio.h>
-#include "fdc-io.h"
-#include "ecc.h"
 #include <sys/errno.h>
 
+#include "ftape.h"
+#include "ecc.h"
+
+/*
+ * Machines that are big-endian should define macro BIG_ENDIAN.
+ * Unfortunately, there doesn't appear to be a standard include
+ * file that works for all OSs.
+ */
+
+#ifdef __sparc__
+# define BIG_ENDIAN
+#endif /* __sparc__ */
+
+#ifdef __hppa
+# define BIG_ENDIAN
+#endif /* __hppa */
+
 #ifdef TEST
 
-# define printk  printf
+# undef TRACE()
+# undef TRACE_()
 # undef TRACE()
 # undef TRACEi()
 # undef TRACElx()
-# define TRACE(l,s1,s2) \
-   {if (ftape_ecc_tracing > (l)) printk(__FILE__ " (%s) - %s.\n", s1, s2);}
-# define TRACEi(l,s1,s2,i) \
-   {if (ftape_ecc_tracing > (l)) printk(__FILE__ " (%s) - %s %d.\n", \
-					s1, s2, i);}
-# define TRACElx(l,s1,s2,lx) \
-   {if (ftape_ecc_tracing > (l)) printk(__FILE__ " (%s) - %s 0x%08lx.\n", \
-				      s1, s2, lx);}
-  int ftape_ecc_tracing = 1;
+# undef TRACE_FUN()
+# undef TRACE_EXIT
+# define printk  printf
+# define TRACE_FUN( level, name) char __fun[] = name
+# define TRACE_EXIT
+# define TRACE_(l,m) \
+{ \
+ if (ftape_ecc_tracing >= (l) && (l) <= TOP_LEVEL) { \
+    printk( "[%03d] " __FILE__ " (%s) - ", (int)ftape_trace_id++, __fun); \
+    m; \
+  } \
+}
+# define TRACE(l,m) TRACE_(l,printk(m".\n"))
+# define TRACEi(l,m,i) TRACE_(l,printk(m" %d.\n",i))
+# define TRACElx(l,m,i) TRACE_(l,printk(m" 0x%08lx.\n",i))
+
+int ftape_ecc_tracing = 1;
+unsigned char ftape_trace_id = 0;
 
 #endif /* TEST */
 
@@ -101,116 +126,44 @@ static const char RCSid[] =
 
 typedef unsigned char Matrix[3][3];
 
-/*
- * This array is a table of powers of r (modulo f(r)):
- */
-static const unsigned char gf_powers[] = {
-    /* first copy for negative indices [-255..-1] */
-    0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80,
-    0x87, 0x89, 0x95, 0xad, 0xdd, 0x3d, 0x7a, 0xf4,
-    0x6f, 0xde, 0x3b, 0x76, 0xec, 0x5f, 0xbe, 0xfb,
-    0x71, 0xe2, 0x43, 0x86, 0x8b, 0x91, 0xa5, 0xcd,
-    0x1d, 0x3a, 0x74, 0xe8, 0x57, 0xae, 0xdb, 0x31,
-    0x62, 0xc4, 0x0f, 0x1e, 0x3c, 0x78, 0xf0, 0x67,
-    0xce, 0x1b, 0x36, 0x6c, 0xd8, 0x37, 0x6e, 0xdc,
-    0x3f, 0x7e, 0xfc, 0x7f, 0xfe, 0x7b, 0xf6, 0x6b,
-    0xd6, 0x2b, 0x56, 0xac, 0xdf, 0x39, 0x72, 0xe4,
-    0x4f, 0x9e, 0xbb, 0xf1, 0x65, 0xca, 0x13, 0x26,
-    0x4c, 0x98, 0xb7, 0xe9, 0x55, 0xaa, 0xd3, 0x21,
-    0x42, 0x84, 0x8f, 0x99, 0xb5, 0xed, 0x5d, 0xba,
-    0xf3, 0x61, 0xc2, 0x03, 0x06, 0x0c, 0x18, 0x30,
-    0x60, 0xc0, 0x07, 0x0e, 0x1c, 0x38, 0x70, 0xe0,
-    0x47, 0x8e, 0x9b, 0xb1, 0xe5, 0x4d, 0x9a, 0xb3,
-    0xe1, 0x45, 0x8a, 0x93, 0xa1, 0xc5, 0x0d, 0x1a,
-    0x34, 0x68, 0xd0, 0x27, 0x4e, 0x9c, 0xbf, 0xf9,
-    0x75, 0xea, 0x53, 0xa6, 0xcb, 0x11, 0x22, 0x44,
-    0x88, 0x97, 0xa9, 0xd5, 0x2d, 0x5a, 0xb4, 0xef,
-    0x59, 0xb2, 0xe3, 0x41, 0x82, 0x83, 0x81, 0x85,
-    0x8d, 0x9d, 0xbd, 0xfd, 0x7d, 0xfa, 0x73, 0xe6,
-    0x4b, 0x96, 0xab, 0xd1, 0x25, 0x4a, 0x94, 0xaf,
-    0xd9, 0x35, 0x6a, 0xd4, 0x2f, 0x5e, 0xbc, 0xff,
-    0x79, 0xf2, 0x63, 0xc6, 0x0b, 0x16, 0x2c, 0x58,
-    0xb0, 0xe7, 0x49, 0x92, 0xa3, 0xc1, 0x05, 0x0a,
-    0x14, 0x28, 0x50, 0xa0, 0xc7, 0x09, 0x12, 0x24,
-    0x48, 0x90, 0xa7, 0xc9, 0x15, 0x2a, 0x54, 0xa8,
-    0xd7, 0x29, 0x52, 0xa4, 0xcf, 0x19, 0x32, 0x64,
-    0xc8, 0x17, 0x2e, 0x5c, 0xb8, 0xf7, 0x69, 0xd2,
-    0x23, 0x46, 0x8c, 0x9f, 0xb9, 0xf5, 0x6d, 0xda,
-    0x33, 0x66, 0xcc, 0x1f, 0x3e, 0x7c, 0xf8, 0x77,
-    0xee, 0x5b, 0xb6, 0xeb, 0x51, 0xa2, 0xc3,
-    /* second copy for indices in range [0..254] */
-    0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80,
-    0x87, 0x89, 0x95, 0xad, 0xdd, 0x3d, 0x7a, 0xf4,
-    0x6f, 0xde, 0x3b, 0x76, 0xec, 0x5f, 0xbe, 0xfb,
-    0x71, 0xe2, 0x43, 0x86, 0x8b, 0x91, 0xa5, 0xcd,
-    0x1d, 0x3a, 0x74, 0xe8, 0x57, 0xae, 0xdb, 0x31,
-    0x62, 0xc4, 0x0f, 0x1e, 0x3c, 0x78, 0xf0, 0x67,
-    0xce, 0x1b, 0x36, 0x6c, 0xd8, 0x37, 0x6e, 0xdc,
-    0x3f, 0x7e, 0xfc, 0x7f, 0xfe, 0x7b, 0xf6, 0x6b,
-    0xd6, 0x2b, 0x56, 0xac, 0xdf, 0x39, 0x72, 0xe4,
-    0x4f, 0x9e, 0xbb, 0xf1, 0x65, 0xca, 0x13, 0x26,
-    0x4c, 0x98, 0xb7, 0xe9, 0x55, 0xaa, 0xd3, 0x21,
-    0x42, 0x84, 0x8f, 0x99, 0xb5, 0xed, 0x5d, 0xba,
-    0xf3, 0x61, 0xc2, 0x03, 0x06, 0x0c, 0x18, 0x30,
-    0x60, 0xc0, 0x07, 0x0e, 0x1c, 0x38, 0x70, 0xe0,
-    0x47, 0x8e, 0x9b, 0xb1, 0xe5, 0x4d, 0x9a, 0xb3,
-    0xe1, 0x45, 0x8a, 0x93, 0xa1, 0xc5, 0x0d, 0x1a,
-    0x34, 0x68, 0xd0, 0x27, 0x4e, 0x9c, 0xbf, 0xf9,
-    0x75, 0xea, 0x53, 0xa6, 0xcb, 0x11, 0x22, 0x44,
-    0x88, 0x97, 0xa9, 0xd5, 0x2d, 0x5a, 0xb4, 0xef,
-    0x59, 0xb2, 0xe3, 0x41, 0x82, 0x83, 0x81, 0x85,
-    0x8d, 0x9d, 0xbd, 0xfd, 0x7d, 0xfa, 0x73, 0xe6,
-    0x4b, 0x96, 0xab, 0xd1, 0x25, 0x4a, 0x94, 0xaf,
-    0xd9, 0x35, 0x6a, 0xd4, 0x2f, 0x5e, 0xbc, 0xff,
-    0x79, 0xf2, 0x63, 0xc6, 0x0b, 0x16, 0x2c, 0x58,
-    0xb0, 0xe7, 0x49, 0x92, 0xa3, 0xc1, 0x05, 0x0a,
-    0x14, 0x28, 0x50, 0xa0, 0xc7, 0x09, 0x12, 0x24,
-    0x48, 0x90, 0xa7, 0xc9, 0x15, 0x2a, 0x54, 0xa8,
-    0xd7, 0x29, 0x52, 0xa4, 0xcf, 0x19, 0x32, 0x64,
-    0xc8, 0x17, 0x2e, 0x5c, 0xb8, 0xf7, 0x69, 0xd2,
-    0x23, 0x46, 0x8c, 0x9f, 0xb9, 0xf5, 0x6d, 0xda,
-    0x33, 0x66, 0xcc, 0x1f, 0x3e, 0x7c, 0xf8, 0x77,
-    0xee, 0x5b, 0xb6, 0xeb, 0x51, 0xa2, 0xc3,
-    /* third copy for indices in range [255..510] */
-    0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80,
-    0x87, 0x89, 0x95, 0xad, 0xdd, 0x3d, 0x7a, 0xf4,
-    0x6f, 0xde, 0x3b, 0x76, 0xec, 0x5f, 0xbe, 0xfb,
-    0x71, 0xe2, 0x43, 0x86, 0x8b, 0x91, 0xa5, 0xcd,
-    0x1d, 0x3a, 0x74, 0xe8, 0x57, 0xae, 0xdb, 0x31,
-    0x62, 0xc4, 0x0f, 0x1e, 0x3c, 0x78, 0xf0, 0x67,
-    0xce, 0x1b, 0x36, 0x6c, 0xd8, 0x37, 0x6e, 0xdc,
-    0x3f, 0x7e, 0xfc, 0x7f, 0xfe, 0x7b, 0xf6, 0x6b,
-    0xd6, 0x2b, 0x56, 0xac, 0xdf, 0x39, 0x72, 0xe4,
-    0x4f, 0x9e, 0xbb, 0xf1, 0x65, 0xca, 0x13, 0x26,
-    0x4c, 0x98, 0xb7, 0xe9, 0x55, 0xaa, 0xd3, 0x21,
-    0x42, 0x84, 0x8f, 0x99, 0xb5, 0xed, 0x5d, 0xba,
-    0xf3, 0x61, 0xc2, 0x03, 0x06, 0x0c, 0x18, 0x30,
-    0x60, 0xc0, 0x07, 0x0e, 0x1c, 0x38, 0x70, 0xe0,
-    0x47, 0x8e, 0x9b, 0xb1, 0xe5, 0x4d, 0x9a, 0xb3,
-    0xe1, 0x45, 0x8a, 0x93, 0xa1, 0xc5, 0x0d, 0x1a,
-    0x34, 0x68, 0xd0, 0x27, 0x4e, 0x9c, 0xbf, 0xf9,
-    0x75, 0xea, 0x53, 0xa6, 0xcb, 0x11, 0x22, 0x44,
-    0x88, 0x97, 0xa9, 0xd5, 0x2d, 0x5a, 0xb4, 0xef,
-    0x59, 0xb2, 0xe3, 0x41, 0x82, 0x83, 0x81, 0x85,
-    0x8d, 0x9d, 0xbd, 0xfd, 0x7d, 0xfa, 0x73, 0xe6,
-    0x4b, 0x96, 0xab, 0xd1, 0x25, 0x4a, 0x94, 0xaf,
-    0xd9, 0x35, 0x6a, 0xd4, 0x2f, 0x5e, 0xbc, 0xff,
-    0x79, 0xf2, 0x63, 0xc6, 0x0b, 0x16, 0x2c, 0x58,
-    0xb0, 0xe7, 0x49, 0x92, 0xa3, 0xc1, 0x05, 0x0a,
-    0x14, 0x28, 0x50, 0xa0, 0xc7, 0x09, 0x12, 0x24,
-    0x48, 0x90, 0xa7, 0xc9, 0x15, 0x2a, 0x54, 0xa8,
-    0xd7, 0x29, 0x52, 0xa4, 0xcf, 0x19, 0x32, 0x64,
-    0xc8, 0x17, 0x2e, 0x5c, 0xb8, 0xf7, 0x69, 0xd2,
-    0x23, 0x46, 0x8c, 0x9f, 0xb9, 0xf5, 0x6d, 0xda,
-    0x33, 0x66, 0xcc, 0x1f, 0x3e, 0x7c, 0xf8, 0x77,
-    0xee, 0x5b, 0xb6, 0xeb, 0x51, 0xa2, 0xc3,
-};
-
 /*
  * gfpow[] is defined such that gfpow[i] returns r^i if
- * i is in the range [-255..509].
+ * i is in the range [0..255].
  */
-static const unsigned char *const gfpow = &gf_powers[255];
+static const unsigned char gfpow[] = {
+    0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80,
+    0x87, 0x89, 0x95, 0xad, 0xdd, 0x3d, 0x7a, 0xf4,
+    0x6f, 0xde, 0x3b, 0x76, 0xec, 0x5f, 0xbe, 0xfb,
+    0x71, 0xe2, 0x43, 0x86, 0x8b, 0x91, 0xa5, 0xcd,
+    0x1d, 0x3a, 0x74, 0xe8, 0x57, 0xae, 0xdb, 0x31,
+    0x62, 0xc4, 0x0f, 0x1e, 0x3c, 0x78, 0xf0, 0x67,
+    0xce, 0x1b, 0x36, 0x6c, 0xd8, 0x37, 0x6e, 0xdc,
+    0x3f, 0x7e, 0xfc, 0x7f, 0xfe, 0x7b, 0xf6, 0x6b,
+    0xd6, 0x2b, 0x56, 0xac, 0xdf, 0x39, 0x72, 0xe4,
+    0x4f, 0x9e, 0xbb, 0xf1, 0x65, 0xca, 0x13, 0x26,
+    0x4c, 0x98, 0xb7, 0xe9, 0x55, 0xaa, 0xd3, 0x21,
+    0x42, 0x84, 0x8f, 0x99, 0xb5, 0xed, 0x5d, 0xba,
+    0xf3, 0x61, 0xc2, 0x03, 0x06, 0x0c, 0x18, 0x30,
+    0x60, 0xc0, 0x07, 0x0e, 0x1c, 0x38, 0x70, 0xe0,
+    0x47, 0x8e, 0x9b, 0xb1, 0xe5, 0x4d, 0x9a, 0xb3,
+    0xe1, 0x45, 0x8a, 0x93, 0xa1, 0xc5, 0x0d, 0x1a,
+    0x34, 0x68, 0xd0, 0x27, 0x4e, 0x9c, 0xbf, 0xf9,
+    0x75, 0xea, 0x53, 0xa6, 0xcb, 0x11, 0x22, 0x44,
+    0x88, 0x97, 0xa9, 0xd5, 0x2d, 0x5a, 0xb4, 0xef,
+    0x59, 0xb2, 0xe3, 0x41, 0x82, 0x83, 0x81, 0x85,
+    0x8d, 0x9d, 0xbd, 0xfd, 0x7d, 0xfa, 0x73, 0xe6,
+    0x4b, 0x96, 0xab, 0xd1, 0x25, 0x4a, 0x94, 0xaf,
+    0xd9, 0x35, 0x6a, 0xd4, 0x2f, 0x5e, 0xbc, 0xff,
+    0x79, 0xf2, 0x63, 0xc6, 0x0b, 0x16, 0x2c, 0x58,
+    0xb0, 0xe7, 0x49, 0x92, 0xa3, 0xc1, 0x05, 0x0a,
+    0x14, 0x28, 0x50, 0xa0, 0xc7, 0x09, 0x12, 0x24,
+    0x48, 0x90, 0xa7, 0xc9, 0x15, 0x2a, 0x54, 0xa8,
+    0xd7, 0x29, 0x52, 0xa4, 0xcf, 0x19, 0x32, 0x64,
+    0xc8, 0x17, 0x2e, 0x5c, 0xb8, 0xf7, 0x69, 0xd2,
+    0x23, 0x46, 0x8c, 0x9f, 0xb9, 0xf5, 0x6d, 0xda,
+    0x33, 0x66, 0xcc, 0x1f, 0x3e, 0x7c, 0xf8, 0x77,
+    0xee, 0x5b, 0xb6, 0xeb, 0x51, 0xa2, 0xc3, 0x01
+};
 
 /*
  * This is a log table.  That is, gflog[r^i] returns i (modulo f(r)).
@@ -248,7 +201,7 @@ static const unsigned char gflog[256] = {
     0x6f, 0x78, 0x19, 0x9a, 0x47, 0x74, 0xa7, 0xc1,
     0x23, 0x53, 0x89, 0xfb, 0x14, 0x5d, 0xf8, 0x97,
     0x2e, 0x4b, 0xb9, 0x60, 0x0f, 0xed, 0x3e, 0xe5,
-    0xf6, 0x87, 0xa5, 0x17, 0x3a, 0xa3, 0x3c, 0xb7,
+    0xf6, 0x87, 0xa5, 0x17, 0x3a, 0xa3, 0x3c, 0xb7
 };
 
 /*
@@ -292,6 +245,24 @@ static const unsigned char gfmul_c0[256] = {
 };
 
 
+/*
+ * Returns V modulo 255 provided V is in the range -255,-254,...,509.
+ */
+static inline unsigned char
+mod255(int v)
+{
+    if (v > 0) {
+	if (v < 255) {
+	    return v;
+	} else {
+	    return v - 255;
+	} /* if */
+    } else {
+	return v + 255;
+    } /* if */
+} /* mod255 */
+
+
 /*
  * Add two numbers in the field.  Addition in this field is
  * equivalent to a bit-wise exclusive OR operation---subtraction
@@ -304,6 +275,17 @@ gfadd(unsigned char a, unsigned char b)
 } /* gfadd */
 
 
+/*
+ * Add two vectors of numbers in the field.  Each byte in A and B get
+ * added individually.
+ */
+static inline unsigned long
+gfadd_long(unsigned long a, unsigned long b)
+{
+    return a ^ b;
+} /* gfadd_long */
+
+
 /*
  * Multiply two numbers in the field:
  */
@@ -311,7 +293,7 @@ static inline unsigned char
 gfmul(unsigned char a, unsigned char b)
 {
     if (a && b) {
-	return gfpow[(unsigned int)gflog[a] + gflog[b]];
+	return gfpow[mod255(gflog[a] + gflog[b])];
     } else {
 	return 0;
     } /* if */
@@ -326,27 +308,67 @@ static inline unsigned char
 gfmul_exp(unsigned char a, int b)
 {
     if (a) {
-	return gfpow[gflog[a] + b];
+	return gfpow[mod255(gflog[a] + b)];
     } else {
 	return 0;
     } /* if */
 } /* gfmul_exp */
 
 
+/*
+ * Just like gfmul_exp, except that A is a vector of numbers.  That is,
+ * each byte in A gets multiplied by gfpow[mod255(B)].
+ */
+static inline unsigned long
+gfmul_exp_long(unsigned long a, int b)
+{
+    TRACE_FUN( 8, "gfmul_exp_long");
+    unsigned char t;
+
+    if (sizeof(long) == 4) {
+        TRACE_EXIT;
+	return
+      ((t = a>>24&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<<24):0) |
+      ((t = a>>16&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<<16):0) |
+      ((t = a>> 8&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<< 8):0) |
+      ((t = a>> 0&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<< 0):0);
+#ifndef LINUX
+    } else if (sizeof(long) == 8) {
+        TRACE_EXIT;
+	return
+      ((t = a>>56&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<<56):0) |
+      ((t = a>>48&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<<48):0) |
+      ((t = a>>40&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<<40):0) |
+      ((t = a>>32&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<<32):0) |
+      ((t = a>>24&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<<24):0) |
+      ((t = a>>16&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<<16):0) |
+      ((t = a>> 8&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<< 8):0) |
+      ((t = a>> 0&0xff) ? (((unsigned long)gfpow[mod255(gflog[t]+b)])<< 0):0);
+#endif
+    } else {
+	TRACEx1( 1, "Error: size of long is %d bytes", (int) sizeof(long));
+    } /* if */
+    TRACE_EXIT;
+    return -1;
+} /* gfmul_exp_long */
+
+
 /*
  * Divide two numbers in the field.  Returns a/b (modulo f(x)).
  */
 static inline unsigned char
 gfdiv(unsigned char a, unsigned char b)
 {
+    TRACE_FUN( 8, "gfdiv");
     if (!b) {
-	TRACE(-1, "gfdiv", "Error: division by zero");
+	TRACE(-1, "Error: division by zero");
 	return 0xff;
     } else if (a == 0) {
 	return 0;
     } else {
-	return gfpow[(int) gflog[a] - gflog[b]];
+	return gfpow[mod255(gflog[a] - gflog[b])];
     } /* if */
+    TRACE_EXIT;
 } /* gfdiv */
 
 
@@ -355,58 +377,61 @@ gfdiv(unsigned char a, unsigned char b)
  * linear system that needs to be solved to determine the error
  * magnitudes.  The first deals with matrices of rank 3, while the
  * second deals with matrices of rank 2.  The error indices are passed
- * in arguments L0,..,L2 (0=first sector, 31=last sector).
+ * in arguments L0,..,L2 (0=first sector, 31=last sector).  The
+ * error indices must be sorted in ascending order, i.e., L0<L1<L2.
  *
  * The linear system that needs to be solved for the error
  * magnitudes is A * b = s, where s is the known vector of
  * syndromes, b is the vector of error magnitudes and A in
  * the ORDER=3 case:
  *
- *    A_3 = {{1/r^LOC[0], 1/r^LOC[1], 1/r^LOC[2]},
- *	    {	       1,	   1,	       1},
- *	    {   r^LOC[0],    r^LOC[1],	r^LOC[2]}}
+ *    A_3 = {{1/r^L[0], 1/r^L[1], 1/r^L[2]},
+ *	    {	     1,	       1,	 1},
+ *	    {	r^L[0],	  r^L[1],   r^L[2]}}
  */
 static inline int
 gfinv3(unsigned char l0, unsigned char l1, unsigned char l2, Matrix Ainv)
 {
+    TRACE_FUN( 8, "gfinv3");
     unsigned char det;
     unsigned char t20, t10, t21, t12, t01, t02;
     int log_det;
 
     /* compute some intermediate results: */
-    t20 = gfpow[(int) l2 - l0];		/* t20 = r^l2/r^l0 */
-    t10 = gfpow[(int) l1 - l0];		/* t10 = r^l1/r^l0 */
-    t21 = gfpow[(int) l2 - l1];		/* t21 = r^l2/r^l1 */
-    t12 = gfpow[(int) l1 - l2];		/* t12 = r^l1/r^l2 */
-    t01 = gfpow[(int) l0 - l1];		/* t01 = r^l0/r^l1 */
-    t02 = gfpow[(int) l0 - l2];		/* t02 = r^l0/r^l2 */
+    t20 = gfpow[l2 - l0];		/* t20 = r^l2/r^l0 */
+    t10 = gfpow[l1 - l0];		/* t10 = r^l1/r^l0 */
+    t21 = gfpow[l2 - l1];		/* t21 = r^l2/r^l1 */
+    t12 = gfpow[l1 - l2 + 255];		/* t12 = r^l1/r^l2 */
+    t01 = gfpow[l0 - l1 + 255];		/* t01 = r^l0/r^l1 */
+    t02 = gfpow[l0 - l2 + 255];		/* t02 = r^l0/r^l2 */
     /*
      * Calculate the determinant of matrix A_3^-1 (sometimes called
      * the Vandermonde determinant):
      */
     det = gfadd(t20, gfadd(t10, gfadd(t21, gfadd(t12, gfadd(t01, t02)))));
     if (!det) {
-	TRACE( 1, "gfinv3", "Inversion failed (3 CRC errors, "
-	      ">0 CRC failures)");
+	TRACE( 1, "Inversion failed (3 CRC errors, >0 CRC failures)");
+	TRACE_EXIT;
 	return 0;
     } /* if */
-    log_det = -gflog[det];
+    log_det = 255 - gflog[det];
 
     /*
      * Now, calculate all of the coefficients:
      */
-    Ainv[0][0] = gfmul_exp(gfadd(gfpow[ l1], gfpow[ l2]), log_det);
-    Ainv[0][1] = gfmul_exp(gfadd(       t21,        t12), log_det);
-    Ainv[0][2] = gfmul_exp(gfadd(gfpow[-l1], gfpow[-l2]), log_det);
+    Ainv[0][0] = gfmul_exp(gfadd(gfpow[    l1], gfpow[    l2]), log_det);
+    Ainv[0][1] = gfmul_exp(gfadd(          t21,           t12), log_det);
+    Ainv[0][2] = gfmul_exp(gfadd(gfpow[255-l1], gfpow[255-l2]), log_det);
 
-    Ainv[1][0] = gfmul_exp(gfadd(gfpow[ l0],  gfpow[l2]), log_det);
-    Ainv[1][1] = gfmul_exp(gfadd(       t20,        t02), log_det);
-    Ainv[1][2] = gfmul_exp(gfadd(gfpow[-l0], gfpow[-l2]), log_det);
+    Ainv[1][0] = gfmul_exp(gfadd(gfpow[    l0], gfpow[    l2]), log_det);
+    Ainv[1][1] = gfmul_exp(gfadd(          t20,           t02), log_det);
+    Ainv[1][2] = gfmul_exp(gfadd(gfpow[255-l0], gfpow[255-l2]), log_det);
 
-    Ainv[2][0] = gfmul_exp(gfadd( gfpow[l0],  gfpow[l1]), log_det);
-    Ainv[2][1] = gfmul_exp(gfadd(       t10,        t01), log_det);
-    Ainv[2][2] = gfmul_exp(gfadd(gfpow[-l0], gfpow[-l1]), log_det);
+    Ainv[2][0] = gfmul_exp(gfadd(gfpow[    l0], gfpow[    l1]), log_det);
+    Ainv[2][1] = gfmul_exp(gfadd(          t10,           t01), log_det);
+    Ainv[2][2] = gfmul_exp(gfadd(gfpow[255-l0], gfpow[255-l1]), log_det);
 
+    TRACE_EXIT;
     return 1;
 } /* gfinv3 */
 
@@ -414,36 +439,37 @@ gfinv3(unsigned char l0, unsigned char l1, unsigned char l2, Matrix Ainv)
 static inline int
 gfinv2(unsigned char l0, unsigned char l1, Matrix Ainv)
 {
+    TRACE_FUN( 8, "gfinv2");
     unsigned char det;
     unsigned char t1, t2;
     int log_det;
 
-    t1 = gfpow[(int) -l0];
-    t2 = gfpow[(int) -l1];
+    t1 = gfpow[255 - l0];
+    t2 = gfpow[255 - l1];
     det = gfadd(t1, t2);
     if (!det) {
-	TRACE( 1, "gfinv2", "Inversion failed (2 CRC errors, "
-	      ">0 CRC failures)");
+ 	TRACE( 1, "Inversion failed (2 CRC errors, >0 CRC failures)");
+	TRACE_EXIT;
 	return 0;
     } /* if */
-    log_det = -gflog[det];
+    log_det = 255 - gflog[det];
 
     /*
      * Now, calculate all of the coefficients:
      */
-    Ainv[0][0] = gfmul_exp( 1, log_det);
-    Ainv[0][1] = gfmul_exp(t2, log_det);
+    Ainv[0][0] = Ainv[1][0] = gfpow[log_det];
 
-    Ainv[1][0] = gfmul_exp( 1, log_det);
+    Ainv[0][1] = gfmul_exp(t2, log_det);
     Ainv[1][1] = gfmul_exp(t1, log_det);
 
+    TRACE_EXIT;
     return 1;
 } /* gfinv2 */
 
 
 /*
- * Multiply matrix A by vector s and return result in vector b.
- * m is assumed to be of order nxn, s and b of order nx1.
+ * Multiply matrix A by vector S and return result in vector B.
+ * M is assumed to be of order NxN, S and B of order Nx1.
  */
 static inline void
 gfmat_mul(int n, Matrix A, unsigned char *s, unsigned char *b)
@@ -463,8 +489,8 @@ gfmat_mul(int n, Matrix A, unsigned char *s, unsigned char *b)
 
 
 /*
- * The Reed Solomon ECC codes are computed over the n-th byte of each
- * block, where n=SECTOR_SIZE.  There are up to 29 blocks of data, and
+ * The Reed Solomon ECC codes are computed over the N-th byte of each
+ * block, where N=SECTOR_SIZE.  There are up to 29 blocks of data, and
  * 3 blocks of ECC.  The blocks are stored contiguously in memory.
  * A segment, consequently, is assumed to have at least 4 blocks:
  * one or more data blocks plus three ECC blocks.
@@ -475,41 +501,69 @@ gfmat_mul(int n, Matrix A, unsigned char *s, unsigned char *b)
  *	   is usually called "erasure", the latter "error."
  */
 /*
- * Compute the three parity bytes for a single column of
- * data.  We use a linear feed-back register to do this.
- * The parity bytes P[0], P[STRIDE], P[2*STRIDE] are computed such that:
+ * Compute the parity bytes for C columns of data, where C is the
+ * number of bytes that fit into a long integer.  We use a linear
+ * feed-back register to do this.  The parity bytes P[0], P[STRIDE],
+ * P[2*STRIDE] are computed such that:
  *
- *		x^k * p(x) = m(x) (modulo g(x))
+ *		x^k * p(x) + m(x) = 0 (modulo g(x))
  *
  * where k = NBLOCKS,
  *	 p(x) = P[0] + P[STRIDE]*x + P[2*STRIDE]*x^2, and
- *       m(x) = sum_{i=0}^k DATA[i*SECTOR_SIZE]*x^i.
+ *	 m(x) = sum_{i=0}^k m_i*x^i.
+ *	 m_i  = DATA[i*SECTOR_SIZE]
  */
 static inline void
-set_parity(unsigned char *data, int nblocks, unsigned char *p, int stride)
+set_parity(unsigned long *data, int nblocks, unsigned long *p, int stride)
 {
-    unsigned char p0, p1, p2, t1, t2, *end;
+    TRACE_FUN( 8, "set_parity");
+    unsigned long p0, p1, p2, t1, t2, *end;
 
-    end = data + nblocks * SECTOR_SIZE;
-
-    /*
-     * First iteration is unfolded to avoid having to initialize p0, p1,
-     * and p2 with 0.
-     */
-    p2 = *data;
-    p0 = p1 = gfmul_c0[p2];
-    data += SECTOR_SIZE;
+    end = data + nblocks * (SECTOR_SIZE/sizeof(long));
+    p0 = p1 = p2 = 0;
     while (data < end) {
-	t1 = gfadd(*data, p0);
-	t2 = gfmul_c0[t1];
-	p0 = gfadd(t2, p1);
-	p1 = gfadd(t2, p2);
+	/*
+	 * The new parity bytes p0_i, p1_i, p2_i are computed from the old
+	 * values p0_{i-1}, p1_{i-1}, p2_{i-1} recursively as:
+	 *
+	 *        p0_i = p1_{i-1} + r^105 * (m_{i-1} - p0_{i-1})
+	 *        p1_i = p2_{i-1} + r^105 * (m_{i-1} - p0_{i-1})
+	 *        p2_i =                    (m_{i-1} - p0_{i-1})
+	 *
+	 * With the initial condition: p0_0 = p1_0 = p2_0 = 0.
+	 */
+	t1 = gfadd_long(*data, p0);
+	/*
+	 * Multiply each byte in t1 by 0xc0:
+	 */
+        if (sizeof(long) == 4) {
+            t2 = ((unsigned long) gfmul_c0[t1>>24 & 0xff])<<24 |
+                 ((unsigned long) gfmul_c0[t1>>16 & 0xff])<<16 |
+		 ((unsigned long) gfmul_c0[t1>> 8 & 0xff])<< 8 |
+		 ((unsigned long) gfmul_c0[t1>> 0 & 0xff])<< 0;
+#ifndef LINUX
+        else if (sizeof(long) == 8) {
+	    t2 = ((unsigned long) gfmul_c0[t1>>56 & 0xff])<<56 |
+	         ((unsigned long) gfmul_c0[t1>>48 & 0xff])<<48 |
+		 ((unsigned long) gfmul_c0[t1>>40 & 0xff])<<40 |
+		 ((unsigned long) gfmul_c0[t1>>32 & 0xff])<<32 |
+		 ((unsigned long) gfmul_c0[t1>>24 & 0xff])<<24 |
+		 ((unsigned long) gfmul_c0[t1>>16 & 0xff])<<16 |
+		 ((unsigned long) gfmul_c0[t1>> 8 & 0xff])<< 8 |
+		 ((unsigned long) gfmul_c0[t1>> 0 & 0xff])<< 0;
+#endif
+        } else {
+            TRACEx1( 1, "Error: long is of size %d", (int) sizeof(long));
+	} /* if */
+	p0 = gfadd_long(t2, p1);
+	p1 = gfadd_long(t2, p2);
 	p2 = t1;
-	data += SECTOR_SIZE;
+	data += SECTOR_SIZE/sizeof(long);
     } /* for */
     *p = p0; p += stride;
     *p = p1; p += stride;
     *p = p2;
+    TRACE_EXIT;
 } /* set_parity */
 
 
@@ -517,20 +571,18 @@ set_parity(unsigned char *data, int nblocks, unsigned char *p, int stride)
  * Compute the 3 syndrome values.  DATA should point to the first byte
  * of the column for which the syndromes are desired.  The syndromes
  * are computed over the first NBLOCKS of rows.  The three bytes will be
- * placed in s[0], s[1], and s[2].
+ * placed in S[0], S[1], and S[2].
  *
- * s[i] is the value of the "message" polynomial evaluated at the i-th root
- * of the generator polynomial g(x).  The message polynomial is defined as:
- *
- *	m(x) = sum_{i=0}^{NBLOCKS} DATA[i] * x^i.
+ * S[i] is the value of the "message" polynomial m(x) evaluated at the
+ * i-th root of the generator polynomial g(x).
  *
  * As g(x)=(x-r^-1)(x-1)(x-r^1) we evaluate the message polynomial at
- * x=r^-1 to get s[0], at x=r^0=1 to get s[1], and at x=r to get s[2].
+ * x=r^-1 to get S[0], at x=r^0=1 to get S[1], and at x=r to get S[2].
  * This could be done directly and efficiently via the Horner scheme.
  * However, it would require multiplication tables for the factors
  * r^-1 (0xc3) and r (0x02).  The following scheme does not require
  * any multiplication tables beyond what's needed for set_parity()
- * anyway and is slightly faster if there are no error and slightly
+ * anyway and is slightly faster if there are no errors and slightly
  * slower if there are errors.  The latter is hopefully the infrequent
  * case.
  *
@@ -540,29 +592,30 @@ set_parity(unsigned char *data, int nblocks, unsigned char *p, int stride)
  *	x^k * p(x) = m(x) (modulo g(x)).
  *
  * That is, to evaluate m(r^m), where r^m is a root of g(x), we can
- * simply evaluate (r^m)^k*p(r^m).  Also, notice that p(x)=0 if and
- * only if s(x)=0.  That is, if all parity bytes are 0, we know there
- * is no error in the data and consequently there is no need to
+ * simply evaluate (r^m)^k*p(r^m).  Also, notice that p is 0 if and
+ * only if s is zero.  That is, if all parity bytes are 0, we know
+ * there is no error in the data and consequently there is no need to
  * compute s(x) at all!  In all other cases, we compute s(x) from p(x)
  * by evaluating (r^m)^k*p(r^m) for m=-1, m=0, and m=1.  The p(x)
  * polynomial is evaluated via the Horner scheme.
  */
-static inline int
-compute_syndromes(unsigned char *data, int nblocks, unsigned char *s)
+static int
+compute_syndromes(unsigned long *data, int nblocks, unsigned long *s)
 {
-    unsigned char p[3];
+    unsigned long p[3];
 
     set_parity(data, nblocks, p, 1);
-    if (p[0] || p[1] || p[2]) {
-	s[0] = gfmul_exp(gfadd(p[0],
-			       gfmul_exp(gfadd(p[1], gfmul_exp(p[2], -1)),
-					 -1)),
-			 -nblocks);
-	s[1] = gfadd(gfadd(p[2],  p[1]),  p[0]);
-	s[2] = gfmul_exp(gfadd(p[0],
-			       gfmul_exp(gfadd(p[1], gfmul_exp(p[2],  1)),
-					 1)),
-			 nblocks);
+    if (p[0] | p[1] | p[2]) {
+	/*
+	 * Some of the checked columns do not have a zero syndrome.  For
+	 * simplicity, we compute the syndromes for all columns that we
+	 * have computed the remainders for.
+	 */
+	s[0] = gfmul_exp_long(gfadd_long(p[0], gfmul_exp_long(gfadd_long(p[1],
+		 gfmul_exp_long(p[2], -1)), -1)), -nblocks);
+	s[1] = gfadd_long(gfadd_long(p[2],  p[1]),  p[0]);
+	s[2] = gfmul_exp_long(gfadd_long(p[0], gfmul_exp_long(gfadd_long(p[1],
+                 gfmul_exp_long(p[2],  1)),  1)),  nblocks);
 	return 0;
     } else {
 	return 1;
@@ -571,19 +624,19 @@ compute_syndromes(unsigned char *data, int nblocks, unsigned char *s)
 
 
 /*
- * Correct the block in the column pointed to by DATA.  There
- * are NBAD CRC errors and their indices are in BAD_LOC[0],
- * up to BAD_LOC[NERASURES-1].  If NERASURES>1, Ainv holds
- * the inverse of the matrix of the linear system that needs to
- * be solved to determine the error magnitudes.  S[0], S[1], and
- * S[2] are the syndrome values.  If row j gets corrected, then
- * bit j will be set in CORRECTION_MAP.
+ * Correct the block in the column pointed to by DATA.  There are NBAD
+ * CRC errors and their indices are in BAD_LOC[0], up to
+ * BAD_LOC[NBAD-1].  If NBAD>1, Ainv holds the inverse of the matrix
+ * of the linear system that needs to be solved to determine the error
+ * magnitudes.  S[0], S[1], and S[2] are the syndrome values.  If row
+ * j gets corrected, then bit j will be set in CORRECTION_MAP.
  */
 static inline int
 correct_block(unsigned char *data, int nblocks,
 	      int nbad, int *bad_loc, Matrix Ainv, unsigned char *s,
 	      BAD_SECTOR *correction_map)
 {
+    TRACE_FUN( 8, "correct_block");
     int ncorrected = 0;
     int i;
     unsigned char t1, t2;
@@ -596,15 +649,15 @@ correct_block(unsigned char *data, int nblocks,
 	/* might have a CRC failure: */
 	if (s[0] == 0) {
 	    /* more than one error */
-	    TRACE( 1, "correct_block", "ECC failed "
-		  "(0 CRC errors, >1 CRC failures).");
+	    TRACE( 1, "ECC failed (0 CRC errors, >1 CRC failures).");
+	    TRACE_EXIT;
 	    return -1;
 	} /* if */
 	t1 = gfdiv(s[1], s[0]);
 	if ((bad_loc[nbad++] = gflog[t1]) >= nblocks) {
-	    TRACEi( 1, "correct_block",
-		   "ECC failed (0 CRC errors, >1 CRC failures): "
-		   "attempt to correct data at ", bad_loc[0]);
+	    TRACE( 1, "ECC failed (0 CRC errors, >1 CRC failures): ");
+            TRACEi( 1, "attempt to correct data at ", bad_loc[0]);
+            TRACE_EXIT;
 	    return -1;
 	} /* if */
 	error_mag[0] = s[1];
@@ -618,24 +671,26 @@ correct_block(unsigned char *data, int nblocks,
 	    Ainv[0][0] = gfpow[bad_loc[0]];
 	} else if (t1 == 0 || t2 == 0) {
 	    /* one erasure and more than one error: */
-	    TRACE( 1, "correct_block", "ECC failed (1 erasure, >1 error)");
+ 	    TRACE( 1, "ECC failed (1 erasure, >1 error)");
+            TRACE_EXIT;
 	    return -1;
 	} else {
 	    /* one erasure, one error: */
 	    if ((bad_loc[nbad++] = gflog[gfdiv(t1, t2)]) >= nblocks) {
-		TRACEi( 1, "correct_block",
-		       "ECC failed (1 CRC error, >1 CRC failures): "
-		       "attempt to correct data at", bad_loc[1]);
+                TRACE( 1, "ECC failed (1 CRC errors, >1 CRC failures): ");
+                TRACEi( 1, "attempt to correct data at ", bad_loc[1]);
+                TRACE_EXIT;
 		return -1;
 	    } /* if */
 	    if (!gfinv2(bad_loc[0], bad_loc[1], Ainv)) {
-		/* inversion failed---must have more than one error */
+                /* inversion failed---must have more than one error */
+                TRACE_EXIT;
 		return -1;
 	    } /* if */
 	} /* if */
-
-	/* FALL THROUGH TO ERROR MAGNITUDE COMPUTATION: */
-
+	/*
+	 *  FALL THROUGH TO ERROR MAGNITUDE COMPUTATION:
+	 */
       case 2:
       case 3:
 	/* compute error magnitudes: */
@@ -643,7 +698,8 @@ correct_block(unsigned char *data, int nblocks,
 	break;
 
       default:
-	TRACE( 1, "correct_block", "Internal Error: number of CRC errors > 3");
+	TRACE( 1, "Internal Error: number of CRC errors > 3");
+        TRACE_EXIT;
 	return -1;
     } /* switch */
 
@@ -662,19 +718,21 @@ correct_block(unsigned char *data, int nblocks,
 	    l = bad_loc[i];
 	    dp = &data[l * SECTOR_SIZE];
 	    *dp = gfadd(*dp, e);
-            *correction_map |= 1 << l;
+	    *correction_map |= 1 << l;
 	    ++ncorrected;
 
 	    log_error_mag = gflog[e];
-	    c0 = gfadd(c0, gfpow[log_error_mag - l]);
+	    c0 = gfadd(c0, gfpow[mod255(log_error_mag - l)]);
 	    c1 = gfadd(c1, e);
-	    c2 = gfadd(c2, gfpow[log_error_mag + l]);
+	    c2 = gfadd(c2, gfpow[mod255(log_error_mag + l)]);
 	} /* if */
     } /* for */
     if (c0 || c1 || c2) {
-	TRACE( 1, "correct_block", "ECC self-check failed, too many errors");
+	TRACE( 1, "ECC self-check failed, too many errors");
+        TRACE_EXIT;
 	return -1;
     } /* if */
+    TRACE_EXIT;
     return ncorrected;
 } /* correct_block */
 
@@ -685,15 +743,17 @@ correct_block(unsigned char *data, int nblocks,
  * Perform a sanity check on the computed parity bytes:
  */
 static int
-sanity_check(unsigned char *data, int nblocks)
+sanity_check(unsigned long *data, int nblocks)
 {
-    unsigned char s[3];
+    TRACE_FUN( 8, "sanity_check");
+    unsigned long s[3];
 
     if (!compute_syndromes(data, nblocks, s)) {
-	TRACE(-1, "sanity_check",
-	      "Internal Error: syndrome self-check failed");
+	TRACE(-1, "Internal Error: syndrome self-check failed");
+        TRACE_EXIT;
 	return 0;
     } /* if */
+    TRACE_EXIT;
     return 1;
 } /* sanity_check */
 
@@ -711,11 +771,12 @@ ecc_set_segment_parity(struct memory_segment* mseg)
     unsigned char *parity_bytes;
 
     parity_bytes = &mseg->data[(mseg->blocks - 3) * SECTOR_SIZE];
-    for (i = 0; i < SECTOR_SIZE; ++i) {
-	set_parity(&mseg->data[i], mseg->blocks - 3,
-		   &parity_bytes[i], SECTOR_SIZE);
+    for (i = 0; i < SECTOR_SIZE; i += sizeof(long)) {
+	set_parity((unsigned long*) &mseg->data[i], mseg->blocks - 3,
+		   (unsigned long*) &parity_bytes[i],
+		   SECTOR_SIZE/sizeof(long));
 #ifdef ECC_PARANOID
-	if (!sanity_check(&mseg->data[i], mseg->blocks)) {
+	if (!sanity_check((unsigned long*)&mseg->data[i], mseg->blocks)) {
 	    return -1;
 	} /* if */
 #endif /* ECC_PARANOID */
@@ -731,24 +792,29 @@ ecc_set_segment_parity(struct memory_segment* mseg)
 int
 ecc_correct_data(struct memory_segment* mseg)
 {
+    TRACE_FUN( 5, "ecc_correct_data");
     int col, i, result;
     int ncorrected = 0;
     int nerasures = 0;		/* # of erasures (CRC errors) */
     int erasure_loc[3];		/* erasure locations */
+    unsigned long ss[3];
     unsigned char s[3];
     Matrix Ainv;
 
     mseg->corrected = 0;
 
     /* find first column that has non-zero syndromes: */
-    for (col = 0; col < SECTOR_SIZE; ++col) {
-	if (!compute_syndromes(&mseg->data[col], mseg->blocks, s)) {
+    for (col = 0; col < SECTOR_SIZE; col += sizeof(long)) {
+	if (!compute_syndromes((unsigned long*)&mseg->data[col],
+			       mseg->blocks, ss))
+	{
 	    /* something is wrong---have to fix things */
 	    break;
 	} /* if */
     } /* for */
     if (col >= SECTOR_SIZE) {
 	/* all syndromes are ok, therefore nothing to correct */
+        TRACE_EXIT;
 	return ECC_OK;
     } /* if */
 
@@ -758,8 +824,8 @@ ecc_correct_data(struct memory_segment* mseg)
 	    if (BAD_CHECK(mseg->read_bad, i)) {
 		if (nerasures >= 3) {
 		    /* this is too much for ECC */
-		    TRACE( 1, "ecc_correct_data",
-			  "ECC failed (>3 CRC errors)");
+		    TRACE( 1, "ECC failed (>3 CRC errors)");
+                    TRACE_EXIT;
 		    return ECC_FAILED;
 		} /* if */
 		erasure_loc[nerasures++] = i;
@@ -773,11 +839,13 @@ ecc_correct_data(struct memory_segment* mseg)
     switch (nerasures) {
       case 2:
 	if (!gfinv2(erasure_loc[0], erasure_loc[1], Ainv)) {
-	    return ECC_FAILED;
+            TRACE_EXIT;
+            return ECC_FAILED;
 	} /* if */
 	break;
       case 3:
 	if (!gfinv3(erasure_loc[0], erasure_loc[1], erasure_loc[2], Ainv)) {
+            TRACE_EXIT;
 	    return ECC_FAILED;
 	} /* if */
 	break;
@@ -787,34 +855,50 @@ ecc_correct_data(struct memory_segment* mseg)
     } /* switch */
 
     do {
-	result = correct_block(&mseg->data[col], mseg->blocks,
-			       nerasures, erasure_loc, Ainv, s,
-			       &mseg->corrected);
-	if (result < 0) {
-	    return ECC_FAILED;
-	} /* if */
-	ncorrected += result;
+	for (i = 0; i < sizeof(long); ++i) {
+	    s[0] = ss[0]; s[1] = ss[1]; s[2] = ss[2];
+	    if (s[0] | s[1] | s[2]) {
+#ifdef BIG_ENDIAN
+		result = correct_block(&mseg->data[col+sizeof(long)-1-i],
+				       mseg->blocks,
+				       nerasures, erasure_loc, Ainv, s,
+				       &mseg->corrected);
+#else
+		result = correct_block(&mseg->data[col+i], mseg->blocks,
+				       nerasures, erasure_loc, Ainv, s,
+				       &mseg->corrected);
+#endif
+		if (result < 0) {
+                    TRACE_EXIT;
+		    return ECC_FAILED;
+		} /* if */
+		ncorrected += result;
+	    } /* if */
+	    ss[0] >>= 8; ss[1] >>= 8; ss[2] >>= 8;
+	} /* for */
 
 #ifdef ECC_SANITY_CHECK
-	if (result > 0 && !sanity_check(&mseg->data[col], mseg->blocks)) {
+	if (!sanity_check((unsigned long*) &mseg->data[col], mseg->blocks)) {
+            TRACE_EXIT;
 	    return ECC_FAILED;
 	} /* if */
 #endif /* ECC_SANITY_CHECK */
 
 	/* find next column with non-zero syndromes: */
-	while (++col < SECTOR_SIZE) {
-	    if (!compute_syndromes(&mseg->data[col], mseg->blocks, s)) {
+	while ((col += sizeof(long)) < SECTOR_SIZE) {
+	    if (!compute_syndromes((unsigned long*) &mseg->data[col],
+				   mseg->blocks, ss))
+	    {
 		/* something is wrong---have to fix things */
 		break;
 	    } /* if */
 	} /* while */
     } while (col < SECTOR_SIZE);
     if (ncorrected && nerasures == 0) {
-	TRACE( 2, "ecc_correct_data",
-	      "block contained error not caught by CRC");
+	TRACE( 2, "block contained error not caught by CRC");
     } /* if */
-    TRACEi((ncorrected > 0) ? 4 : 8, "ecc_correct_data",
-	   "number of corrections:", ncorrected);
+    TRACEi((ncorrected > 0) ? 4 : 8, "number of corrections:", ncorrected);
+    TRACE_EXIT;
     return ncorrected ? ECC_CORRECTED : ECC_OK;
 } /* ecc_correct_data */
 
